package adt2prism;

import java.util.ArrayList;

import adt2prism.ADConstants.Refinement;
import adt2prism.ADConstants.Player;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

/**
 *
 * @author zaas
 * A node for representing attack-defence tree.
 * The tree is a pointer to the root.
 */
public class ADNode {
    
    private ArrayList<ADNode> children;
    private double cost;
    private double probability;
    private String label;
    private Refinement type;
    private Player owner;
    private ADNode parent;
    
    //Constructor for a node with label, type and owner
    ADNode(String l, Refinement t, Player o, ADNode p){
        this.label = l;
        this.type = t;
        this.owner = o;
        this.parent = p;
    }

    //add child to the node
    public boolean addChild(ADNode child){
        if (this.children==null)
            this.children = new ArrayList<>();
        return this.children.add(child);
    }
    
    public boolean isRoot() {
        return (parent == null);
    }
    
    //set cost value
    public void setCost(double cost) {
        this.cost = cost;
    }

    //set probability value
    public void setProbability(double probability) {
        this.probability = probability;
    }

    //set label of the node
    public void setLabel(String label) {
        this.label = label;
    }

    //set type of the node
    public void setType(Refinement type) {
        this.type = type;
    }

    //set owner of the node
    public void setOwner(Player owner) {
        this.owner = owner;
    }

    //get children of the node
    public ArrayList<ADNode> getChildren() {
        return children;
    }

    //get cost of the node
    public double getCost() {
        return cost;
    }

    //get probability of the node
    public double getProbability() {
        return probability;
    }

    //get label of the node
    public String getLabel() {
        return label;
    }

    //get type of the node
    public Refinement getType() {
        return type;
    }

    //get owner of the node
    public Player getOwner() {
        return owner;
    }
    
    //get parent of the node
    public ADNode getParent() {
        return parent;
    }
    
    //get all leaves of the tree rooted at "this"
    public ArrayList<ADNode> getLeaves(ArrayList<ADNode> res){
        if(res==null)
            res = new ArrayList<ADNode>();
        
        if(this.type == Refinement.LEAF)
            res.add(this);
        else
            for(int i=0; i < children.size(); i++)
                children.get(i).getLeaves(res);
        return res;
    }
    
    //get all leaves in the tree owned by player o
    public ArrayList<ADNode> getLeaves(ArrayList<ADNode> res, Player o){
        if(res==null)
           res = new ArrayList<ADNode>();
        
        if(this.type == Refinement.LEAF && this.owner == o)
            res.add(this);
        else if (children != null)
            for(int i=0; i < children.size(); i++)
                children.get(i).getLeaves(res, o);
        //else if (children == null)
            //System.out.println("null" + this.getLabel());
        return res;
    }
    
    // return an arrayList containing all the roots of maximal non-sequential subtrees from left to right
    // Assume that no sequential op occurs under a non-sequential one
    public ArrayList<ADNode> getNonSequentialSubTrees(ArrayList<ADNode> res) {
        if (res==null) 
            res = new ArrayList<ADNode>();
        
        if (this.type != Refinement.SAND && this.type != Refinement.SOR)
            res.add(this);
        else
            for (int i=0; i < children.size(); i++)
                children.get(i).getNonSequentialSubTrees(res);
        return res;
    }
    
    public Hashtable<String,ADNode[]> getLeavesCombinations() {
        ArrayList<ADNode> basicActions = this.getLeaves(null);  
        ADNode[][] combinations = powerset(basicActions.toArray(new ADNode[basicActions.size()]));     
        Hashtable<String, ADNode[]> table = new Hashtable<String, ADNode[]>();
        
        for (int i=0; i<combinations.length; i++) {
            String key = "do_";
      
            for(int j=0; j < combinations[i].length; j++) {
                key += combinations[i][j].getLabel().replace(" ", "").replace("-", "_");
                key += "_";
            }      
            if (key.equals("do_"))
                    key = "do_nothing";
            
            table.put(key, combinations[i]);
        } 
        return table;
    }
    
    public Hashtable<String,ADNode[]> getLeavesCombinations(Player p) {
        ArrayList<ADNode> basicActions = this.getLeaves(null, p);  
        ADNode[][] combinations = powerset(basicActions.toArray(new ADNode[basicActions.size()]));     
        Hashtable<String, ADNode[]> table = new Hashtable<String, ADNode[]>();
        
        for (int i=0; i<combinations.length; i++) {
            String key = "do_";
      
            for(int j=0; j < combinations[i].length; j++) {
                key += combinations[i][j].getLabel().replace(" ", "").replace("-", "_");
                key += "_";
            }      
            if (key.equals("do_"))
                    key = "do_nothing";
            
            table.put(key, combinations[i]);
        } 
        return table;
    }
    
    public static ADNode[][] powerset(ADNode[] set) {

    List<ADNode[]> combinationList = new ArrayList<ADNode[]>();
    // Start i at 1, so that we do not include the empty set in the results
        for ( long i = 0; i < Math.pow(2, set.length); i++ ) {
            List<ADNode> nodeList = new ArrayList<ADNode>();
            for ( int j = 0; j < set.length; j++ ) {
                if ( (i & (long) Math.pow(2, j)) > 0 ) {
                    // Include j in set
                    nodeList.add(set[j]);
                }
            }
            combinationList.add(nodeList.toArray(new ADNode[0]));
        }
        return combinationList.toArray(new ADNode[0][0]);
    }

    public String getProbabilityFormula(ADNode[] performedBasicActions) {
        String formula = "";
        
        if (this.type == Refinement.LEAF && this.evaluateBoolean(performedBasicActions))
            formula = "prob_" + this.getLabel().replace(" ", "").replace("-", "_");
        else if (this.type == Refinement.LEAF && !this.evaluateBoolean(performedBasicActions))
            formula = "0";
        
        else if (this.type == Refinement.AND || this.type == Refinement.SAND && this.evaluateBoolean(performedBasicActions)) {
            for (int i=0; i<this.children.size(); i++) {
                if (i<this.children.size()-1 && this.owner != children.get(i).owner)
                    formula += "(1-(" + children.get(i).getProbabilityFormula(performedBasicActions) + "))" + "*";
                else if (i<this.children.size()-1 && this.owner == children.get(i).owner)
                    formula += "(" + children.get(i).getProbabilityFormula(performedBasicActions) + ")" + "*";
                else if (i==this.children.size()-1 && this.owner != children.get(i).owner)
                    formula += "(1-(" + children.get(i).getProbabilityFormula(performedBasicActions) + "))";
                else if (i==this.children.size()-1 && this.owner == children.get(i).owner)
                    formula += "(" + children.get(i).getProbabilityFormula(performedBasicActions) + ")";      
            } 
        } 
        else if (this.type == Refinement.OR || this.type == Refinement.SOR  && this.evaluateBoolean(performedBasicActions)) {
            formula = "1 - (";
            for (int i=0; i<this.children.size(); i++) {
                if (i<this.children.size()-1 && this.owner != children.get(i).owner)
                    formula += "(" + children.get(i).getProbabilityFormula(performedBasicActions) +")" + "*";
                else if (i<this.children.size()-1 && this.owner == children.get(i).owner)
                    formula += "(1-(" + children.get(i).getProbabilityFormula(performedBasicActions) +"))" + "*";
                else if (i==this.children.size()-1 && this.owner != children.get(i).owner)
                    formula += "(" + children.get(i).getProbabilityFormula(performedBasicActions) +")";
                else if (i==this.children.size()-1 && this.owner == children.get(i).owner)
                    formula += "(1-(" + children.get(i).getProbabilityFormula(performedBasicActions) +"))";
            }
            formula += ")";
        }  
        return formula;
    }
    
    public boolean evaluateBoolean(ADNode[] performedBasicActions) {

        if (performedBasicActions.length == 0)
            return false;
        
        if (this.type == Refinement.LEAF) {
            return this.isContained(performedBasicActions);
        }
        else if (this.type == Refinement.AND || this.type == Refinement.SAND) {
            for (int i=0; i<children.size(); i++) {  
               if (children.get(i).owner == this.owner && !children.get(i).evaluateBoolean(performedBasicActions))
                   return false;
               else if (children.get(i).owner != this.owner && children.get(i).evaluateBoolean(performedBasicActions))
                   return false;
            }
            return true;
        }
        else if (this.type == Refinement.OR || this.type == Refinement.SOR) {
            for (int i=0; i<children.size(); i++) {  
               if (children.get(i).owner == this.owner && children.get(i).evaluateBoolean(performedBasicActions))
                   return true;
               else if (children.get(i).owner != this.owner && !children.get(i).evaluateBoolean(performedBasicActions))
                   return true;
            }
            return false;
        }
        return false;
    }
    
    public boolean isContained(ADNode[] nodes) {
       for (int i=0; i<nodes.length; i++) {
                if (nodes[i].getLabel().equals(this.getLabel()))
                    return true;
            }
       return false;
    }
    
    
    @Override
    public String toString() {
        return this.toString(0);
    }
    
    public String toString(int indent) {
    String res = "";
        
        for (int j=0; j<=indent; j++)
            res+= "\t";
        if (this.type != ADConstants.Refinement.LEAF) {
            res += "Node - " + this.type + ", " + this.label + ", " + this.owner +"\n";

            for (int i = 0; i < this.children.size(); i++) {
                res += this.children.get(i).toString(indent+1) + "\n";
            }
        }
        if (this.type == ADConstants.Refinement.LEAF)
            res += "Leaf - " + "cost: " + this.cost + ", probability " + this.probability  + ", " + this.label + ", " + this.owner + "\n";
        
        return res;
    }
    
    // transform attack-defence tree rooted at "this" into a PRISM game
    // consider factoring out a method for each refinement type to improve readability
    public StringBuilder toPRISMGames(){
        StringBuilder sb = new StringBuilder();
        
        //intex for state in MDP
        int z =0;
        
        sb.append("smg").append("\n \n");
        //create parameters for each leaves' probability value
        ArrayList<ADNode> leaves = this.getLeaves(null);
        for(int i=0; i<leaves.size(); i++){
            sb.append("const double prob_");
            sb.append(leaves.get(i).getLabel().replace(" ", "").replace("-", "_"));
            sb.append(" = ");
            sb.append(leaves.get(i).getProbability());
            sb.append("; \n");
        }
        
        //print player "a" (attacker) with his/her actions
        sb.append("\n player a \n");
        sb.append("attacker, ");
        
        //make action label for each sub-set
        ArrayList<ADNode> nonSequentialTrees = this.getNonSequentialSubTrees(null);
        for(int i=0; i<nonSequentialTrees.size(); i++){
            Hashtable<String,ADNode[]> leavesCombination = nonSequentialTrees.get(i).getLeavesCombinations(ADConstants.Player.attacker);
            Set<String> keys = leavesCombination.keySet();
            for(String key : keys){
                sb.append("[").append(key).append(i).append("_att],");
            }
        }
        sb.append("\n");
        for(int i=0; i<nonSequentialTrees.size(); i++){
            sb.append("[prob_phase_").append(i).append("],");
        }
        sb.append("\n [succeed], [fail]");
        sb.append("\n endplayer \n \n");
        
        //print player "d" (defender) with his/her actions
        sb.append("\n player d \n");
        
        //make action label for each sub-set
        for(int i=0; i<nonSequentialTrees.size(); i++){
            Hashtable<String,ADNode[]> leavesCombination = nonSequentialTrees.get(i).getLeavesCombinations(ADConstants.Player.defender);
            Set<String> keys = leavesCombination.keySet();

            int c=0;
            for(String key : keys){
                if (i < nonSequentialTrees.size()-1 || c < keys.size()-1 )
                    sb.append("[").append(key).append(i).append("_def],");
                else
                    sb.append("[").append(key).append(i).append("_def]");
                c++;
            }
        }
        sb.append("\n endplayer \n \n");
        
        //start module
        sb.append("module attacker \n \n");
        sb.append("s:[0..").append(3*nonSequentialTrees.size()+1).append("]; \n \n");
        
        //print parametres for all actions in the tree
        for(int i=0; i<leaves.size(); i++){
            sb.append(leaves.get(i).getLabel().replace(" ", "").replace("-", "_"));
            sb.append(" : bool init false; \n");
        }
        sb.append("\n");
        //print parameters for each phase
        for(int i=0; i<nonSequentialTrees.size(); i++){
            sb.append("success_phase_").append(i).append(" : bool init false; \n");
        }
        sb.append("\n");
        //for each non-sequential sub-tree t_i
        for(int i=0; i<nonSequentialTrees.size(); i++){
            //do defender's nondeterministic actions for t_i
            Hashtable<String,ADNode[]> leavesCombinationDefender = nonSequentialTrees.get(i).getLeavesCombinations(ADConstants.Player.defender);
            Set<String> keysDefender = leavesCombinationDefender.keySet();
            for(String key : keysDefender){
                sb.append("[").append(key).append(i).append("_def] ");
                sb.append("s=").append(z).append(" -> (s'=").append(z+1).append(")");
                ADNode[] values = leavesCombinationDefender.get(key);
                for(int j=0; j<values.length; j++){
                    sb.append("&(").append(values[j].getLabel().replace(" ", "").replace("-", "_")).append("'=true)");
                }
                sb.append(";\n");
            }
            sb.append("\n");
            z++;
            
            //do attacker's nondeterministic actions for t_i
            Hashtable<String,ADNode[]> leavesCombinationAttacker = nonSequentialTrees.get(i).getLeavesCombinations(ADConstants.Player.attacker);
            Set<String> keysAttacker = leavesCombinationAttacker.keySet();
            for(String key : keysAttacker){
                sb.append("[").append(key).append(i).append("_att] ");
                sb.append("s=").append(z).append(" -> (s'=").append(z+1).append(")");
                ADNode[] values = leavesCombinationAttacker.get(key);
                for(int j=0; j<values.length; j++){
                    sb.append("&(").append(values[j].getLabel().replace(" ", "").replace("-", "_")).append("'=true)");
                }
                sb.append(";\n");
            }
            sb.append("\n");
            z++;
            
            //do probabilistic transitions for t_i
            ArrayList<ADNode> currentleaves = nonSequentialTrees.get(i).getLeaves(null);
            Hashtable<String,ADNode[]> leavesCombination = nonSequentialTrees.get(i).getLeavesCombinations();
            
            Set<String> keys = leavesCombination.keySet();
            // for(String key : keys){
            //if there exist a previous  phase
            if(i > 0){
                  
                //if the parent of the current non-sequential sub-tree is a conjunction
                if(nonSequentialTrees.get(i).parent.type == Refinement.SAND){

                    //for each combination of basic actions and success_phase_i-1 = TRUE
                    for(String key : keys){
                        ADNode[] values = leavesCombination.get(key);
                        sb.append("[prob_phase_").append(i).append("] ");
                        sb.append("s=").append(z);
                        sb.append("& success_phase_").append(i-1).append("=true");
                        //print the truth values of basic actions
                        for(int j=0; j<currentleaves.size(); j++){
                            if (currentleaves.get(j).isContained(values)) {
                                // current basic action is performed, i.e., true
                                sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=true");
                            }
                            else 
                                //current basic action is not performed, i.e., false
                                sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=false");
                        }
                        sb.append(" -> ");
                        //print probabilistic transitions
                        //if the current tree consists of one node 
                        if(nonSequentialTrees.get(i).getType() == Refinement.LEAF){
                            //if the owner of the node is different from the owner of the parent 
                            if(nonSequentialTrees.get(i).getOwner() != nonSequentialTrees.get(i).getParent().getOwner()){
                                //if there exists a performed action
                                if(values.length >0){
                                    //probability of success, i.e. success_phase_i '= true
                                    sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                    //probability of failure, i.e. success_phase_i '= false
                                    sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }
                                //if the set of performed acions is empty
                                else{
                                    //failure with probability 1, i.e. success_phase_i '= false
                                    sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true);");
                                }
                            }
                            //if the owner of the node is the same as the owner of the parent 
                            else{
                                //if there exist a performed actions
                                if(values.length >0){
                                    //probability of success, i.e. success_phase_i '= true
                                    sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                    //probability of failure, i.e. success_phase_i '= false
                                    sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }
                                //if performed action evaluates to false
                                else{
                                    //failure with probability 1, i.e. success_phase_i '= false
                                    sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }

                            }
                        }
                        //if the current tree consists of more than one node
                        else{
                            //if the set of performed actions is not empty
                            if(values.length >0){
                                //probability of success, i.e. success_phase_i '= true
                                sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                //probability of failure, i.e. success_phase_i '= false
                                sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }
                            //if the set of performed actions is empty
                            else {
                                //failure with probability 1, i.e. success_phase_i '= false
                                sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }
                        }
                        sb.append("\n");
                    }

                    //when success_phase_i-1 = FALSE then all phase_i is false
                    sb.append("[prob_phase_").append(i).append("] ");
                    sb.append("s=").append(z);
                    sb.append("& success_phase_").append(i-1).append("=false");
                    sb.append(" -> ");
                    sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                    sb.append("\n");
                }
                
                //if the parent of the current non-sequential sub-tree is a conjunction
                if(nonSequentialTrees.get(i).parent.type == Refinement.SOR){

                    //for each combination of basic actions and success_phase_i-1 = TRUE
                    for(String key : keys){
                        ADNode[] values = leavesCombination.get(key);
                        sb.append("[prob_phase_").append(i).append("] ");
                        sb.append("s=").append(z);
                        sb.append("& success_phase_").append(i-1).append("=true");
                        //print the truth values of basic actions
                        for(int j=0; j<currentleaves.size(); j++){
                            if (currentleaves.get(j).isContained(values)) {
                                // current basic action is performed, i.e., true
                                sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=true");
                            }
                            else 
                                //current basic action is not performed, i.e., false
                                sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=false");
                        }
                        sb.append(" -> ");
                        //print probabilistic transitions
                        //if the current tree consists of one node 
                        if(nonSequentialTrees.get(i).getType() == Refinement.LEAF){
                            //if the owner of the node is different from the owner of the parent 
                            if(nonSequentialTrees.get(i).getOwner() != nonSequentialTrees.get(i).getParent().getOwner()){
                                //if there exists a performed action
                                if(values.length >0){
                                    //probability of success, i.e. success_phase_i '= true
                                    sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                    //probability of failure, i.e. success_phase_i '= false
                                    sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }
                                //if the set of performed acions is empty
                                else{
                                    //failure with probability 1, i.e. success_phase_i '= false
                                    sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true);");
                                }
                            }
                            //if the owner of the node is the same as the owner of the parent 
                            else{
                                //if there exist a performed actions
                                if(values.length >0){
                                    //probability of success, i.e. success_phase_i '= true
                                    sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                    //probability of failure, i.e. success_phase_i '= false
                                    sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }
                                //if performed action evaluates to false
                                else{
                                    //failure with probability 1, i.e. success_phase_i '= false
                                    sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true);");
                                }

                            }
                        }
                        //if the current tree consists of more than one node
                        else{
                            //if the set of performed actions is not empty
                            if(values.length >0){
                                //probability of success, i.e. success_phase_i '= true
                                sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                //probability of failure, i.e. success_phase_i '= false
                                sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }
                            //if the set of performed actions is empty
                            else {
                                //failure with probability 1, i.e. success_phase_i '= false
                                sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true);");
                            }
                        }
                        sb.append("\n");
                    }

                    //for each combination of basic actions and success_phase_i-1 = FALSE
                    for(String key : keys){
                        ADNode[] values = leavesCombination.get(key);
                        sb.append("[prob_phase_").append(i).append("] ");
                        sb.append("s=").append(z);
                        sb.append("& success_phase_").append(i-1).append("=false");
                        //print the truth values of basic actions
                        for(int j=0; j<currentleaves.size(); j++){
                            if (currentleaves.get(j).isContained(values)) {
                                // current basic action is performed, i.e., true
                                sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=true");
                            }
                            else 
                                //current basic action is not performed, i.e., false
                                sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=false");
                        }
                        sb.append(" -> ");
                        //print probabilistic transitions
                        //if the current tree consists of one node 
                        if(nonSequentialTrees.get(i).getType() == Refinement.LEAF){
                            //if the owner of the node is different from the owner of the parent 
                            if(nonSequentialTrees.get(i).getOwner() != nonSequentialTrees.get(i).getParent().getOwner()){
                                //if there exists a performed action
                                if(values.length >0){
                                    //probability of success, i.e. success_phase_i '= true
                                    sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                    //probability of failure, i.e. success_phase_i '= false
                                    sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }
                                //if the set of performed acions is empty
                                else{
                                    //failure with probability 1, i.e. success_phase_i '= false
                                    sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true);");
                                }
                            }
                            //if the owner of the node is the same as the owner of the parent 
                            else{
                                //if there exist a performed actions
                                if(values.length >0){
                                    //probability of success, i.e. success_phase_i '= true
                                    sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                    //probability of failure, i.e. success_phase_i '= false
                                    sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                    sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }
                                //if performed action evaluates to false
                                else{
                                    //failure with probability 1, i.e. success_phase_i '= false
                                    sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                                }

                            }
                        }
                        //if the current tree consists of more than one node
                        else{
                            //if the set of performed actions is not empty
                            if(values.length >0){
                                //probability of success, i.e. success_phase_i '= true
                                sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                //probability of failure, i.e. success_phase_i '= false
                                sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }
                            //if the set of performed actions is empty
                            else {
                                //failure with probability 1, i.e. success_phase_i '= false
                                sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }
                        }
                        sb.append("\n");
                    }
                }

            }
                
            //if it is the first phase
            else{
                for(String key : keys){
                    //for each combination of basic actions
                    ADNode[] values = leavesCombination.get(key);
                    sb.append("[prob_phase_").append(i).append("] ");
                    sb.append("s=").append(z);

                    //print the truth values of basic actions
                    for(int j=0; j<currentleaves.size(); j++){
                        if (currentleaves.get(j).isContained(values)) {
                            // current basic action is performed, i.e., true
                            sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=true");
                        }
                        else 
                            //current basic action is not performed, i.e., false
                            sb.append("&").append(currentleaves.get(j).getLabel().replace(" ", "").replace("-", "_")).append("=false");
                    }
                    sb.append(" -> ");
                    //print probabilistic transitions
                    //if the current tree consists of one node 
                    if(nonSequentialTrees.get(i).getType() == Refinement.LEAF){
                        //if the owner of the node is different from the owner of the parent 
                        if(nonSequentialTrees.get(i).getOwner() != nonSequentialTrees.get(i).getParent().getOwner()){
                            //if the set of performed actions is not empty
                            if(values.length >0){
                                //probability of success, i.e. success_phase_i '= true
                                sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                //probability of failure, i.e. success_phase_i '= false
                                sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }
                            //if the set of performed actions is empty
                            else{
                                //failure with probability 1, i.e. success_phase_i '= false
                                sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true);");
                            }
                        }
                        //if the owner of the node is the same as the owner of the parent 
                        else{
                            //if the set of performed actions is not empty
                            if(values.length >0){
                                //probability of success, i.e. success_phase_i '= true
                                sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                                //probability of failure, i.e. success_phase_i '= false
                                sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                                sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }
                            //if the set of performed actions is empty
                            else{
                                //failure with probability 1, i.e. success_phase_i '= false
                                sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                            }

                        }
                    }
                    //if the current tree consists of more than one node
                    else{
                        //if the set of performed actions si not empty
                        if(values.length >0){
                            //probability of success, i.e. success_phase_i '= true
                            sb.append(nonSequentialTrees.get(i).getProbabilityFormula(values));
                            sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=true) + ");
                            //probability of failure, i.e. success_phase_i '= false
                            sb.append("1-(").append(nonSequentialTrees.get(i).getProbabilityFormula(values)).append(")");
                            sb.append(":(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                        }
                        //if the set of performed actions is empty
                        else {
                            //failure with probability 1, i.e. success_phase_i '= false
                            sb.append("(s'=").append(z+1).append(")&(success_phase_").append(i).append("'=false);");
                        }
                    }
                    sb.append("\n");
                }
            }

            sb.append("\n");
            
            z++;
        }
        
        //[succeed] actions
        sb.append("[succeed] s=").append(z);
        sb.append("& success_phase_").append(nonSequentialTrees.size()-1).append("=true");
        sb.append(" -> (s'=").append(z+1).append("); \n");
        //[fail] actions
        sb.append("[fail] s=").append(z);
        sb.append("& success_phase_").append(nonSequentialTrees.size()-1).append("=false");
        sb.append(" -> (s'=").append(z+1).append("); \n \n");
        
        //end module
        sb.append("endmodule \n \n");
        
        //success state for an attacker
        sb.append("//success state for an attacker \n");
        sb.append("label \"success\" = s=").append(z+1).append("&success_phase_").append(nonSequentialTrees.size()-1).append("=true; \n");
        sb.append("label \"end\" = s=").append(z+1).append("; \n \n");
        
        //reward cost for each combination of attacker's basic actions
        sb.append("rewards \"cost\" \n");
        for(int i=0; i<nonSequentialTrees.size(); i++){
            Hashtable<String,ADNode[]> leavesCombination = nonSequentialTrees.get(i).getLeavesCombinations(ADConstants.Player.attacker);
            Set<String> keys = leavesCombination.keySet();
            //for each leaves combination create summed cost value
            for(String key : keys){
                int cost = 0;
                ADNode[] values =leavesCombination.get(key);
                if(values.length >0){
                    for(int j=0; j<values.length; j++)
                        cost += values[j].getCost();
                    sb.append("[").append(key).append(i).append("_att] true:").append(cost).append("; \n");
                }
            }
        }
        sb.append("endrewards \n \n");
        
        
        //reward success 
        sb.append("rewards \"success\" \n");
        sb.append("[succeed] true:1; \n");
        sb.append("endrewards \n");
        
        return sb;
    }
    
}
